"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const configs_1 = require("./configs");
var BUTTON_TYPE;
(function (BUTTON_TYPE) {
    BUTTON_TYPE["ARROW_UP"] = "ARROW_UP";
    BUTTON_TYPE["ARROW_DOWN"] = "ARROW_DOWN";
    BUTTON_TYPE["ARROW_LEFT"] = "ARROW_LEFT";
    BUTTON_TYPE["ARROW_RIGHT"] = "ARROW_RIGHT";
    BUTTON_TYPE["HOME"] = "HOME";
    BUTTON_TYPE["START"] = "START";
    BUTTON_TYPE["SELECT"] = "SELECT";
    BUTTON_TYPE["CLUSTER_TOP"] = "CLUSTER_TOP";
    BUTTON_TYPE["CLUSTER_LEFT"] = "CLUSTER_LEFT";
    BUTTON_TYPE["CLUSTER_RIGHT"] = "CLUSTER_RIGHT";
    BUTTON_TYPE["CLUSTER_BOTTOM"] = "CLUSTER_BOTTOM";
    BUTTON_TYPE["BUMPER_TOP_LEFT"] = "BUMPER_TOP_LEFT";
    BUTTON_TYPE["BUMPER_BOTTOM_LEFT"] = "BUMPER_BOTTOM_LEFT";
    BUTTON_TYPE["BUMPER_TOP_RIGHT"] = "BUMPER_TOP_RIGHT";
    BUTTON_TYPE["BUMPER_BOTTOM_RIGHT"] = "BUMPER_BOTTOM_RIGHT";
    BUTTON_TYPE["STICK_PRESS_LEFT"] = "STICK_PRESS_LEFT";
    BUTTON_TYPE["STICK_PRESS_RIGHT"] = "STICK_PRESS_RIGHT";
    BUTTON_TYPE["TOUCHSCREEN"] = "TOUCHSCREEN";
})(BUTTON_TYPE = exports.BUTTON_TYPE || (exports.BUTTON_TYPE = {}));
var STICK_TYPE;
(function (STICK_TYPE) {
    STICK_TYPE["LEFT"] = "LEFT";
    STICK_TYPE["RIGHT"] = "RIGHT";
})(STICK_TYPE = exports.STICK_TYPE || (exports.STICK_TYPE = {}));
var ANALOG_TYPE;
(function (ANALOG_TYPE) {
    ANALOG_TYPE["BUMPER_LEFT"] = "BUMPER_LEFT";
    ANALOG_TYPE["BUMPER_RIGHT"] = "BUMPER_RIGHT";
})(ANALOG_TYPE = exports.ANALOG_TYPE || (exports.ANALOG_TYPE = {}));
var DIRECTION;
(function (DIRECTION) {
    DIRECTION["UP"] = "UP";
    DIRECTION["DOWN"] = "DOWN";
    DIRECTION["LEFT"] = "LEFT";
    DIRECTION["RIGHT"] = "RIGHT";
})(DIRECTION = exports.DIRECTION || (exports.DIRECTION = {}));
var KEY_MODIFIER;
(function (KEY_MODIFIER) {
    KEY_MODIFIER["ALT"] = "ALT";
    KEY_MODIFIER["CTRL"] = "CTRL";
    KEY_MODIFIER["META"] = "META";
    KEY_MODIFIER["SHIFT"] = "SHIFT";
})(KEY_MODIFIER = exports.KEY_MODIFIER || (exports.KEY_MODIFIER = {}));
function lookupButtonIndex(gamepad, arg) {
    const config = configs_1.getGamepadConfig(gamepad);
    // little more complicated because the index could be 0
    if (config) {
        const buttonIndex = config.buttons[arg];
        if (typeof buttonIndex !== 'undefined') {
            return buttonIndex;
        }
    }
    return null;
}
function lookupStickIndex(gamepad, arg) {
    const config = configs_1.getGamepadConfig(gamepad);
    return config && config.sticks[arg] || null;
}
class ControllerGamepad {
    constructor(index) {
        if (!(index >= 0)) {
            throw new Error(`Must provide an index. Or, use the AnyGamepad class`);
        }
        this.index = index;
    }
    isSomethingConnected() {
        return !!this.raw();
    }
    isRecognized() {
        const raw = this.raw();
        return !!raw && configs_1.getGamepadConfig(raw) !== null;
    }
    isButtonPressed(arg) {
        const raw = this.raw();
        if (raw) {
            const index = this.buttonIndex(arg);
            return index !== null && raw.buttons[index].pressed;
        }
        else {
            return false;
        }
    }
    hasButton(arg) {
        return this.buttonIndex(arg) !== null;
    }
    button(arg) {
        return {
            has: () => this.hasButton(arg),
            query: () => this.isButtonPressed(arg)
        };
    }
    getStickValue(arg) {
        const raw = this.raw();
        if (raw) {
            const indexes = this.stickIndex(arg);
            if (indexes) {
                const { xAxis, yAxis } = indexes;
                return { x: raw.axes[xAxis], y: raw.axes[yAxis] };
            }
        }
        return { x: 0, y: 0 };
    }
    getStickDirection(arg) {
        const { x, y } = this.getStickValue(arg);
        const xAbs = Math.abs(x);
        const yAbs = Math.abs(y);
        if (xAbs < .5 && yAbs < .5) {
            return null;
        }
        else if (xAbs > yAbs) {
            return x > 0 ? DIRECTION.LEFT : DIRECTION.RIGHT;
        }
        else {
            return y > 0 ? DIRECTION.DOWN : DIRECTION.UP;
        }
    }
    hasStick(arg) {
        return this.stickIndex(arg) !== null;
    }
    stick(arg) {
        return {
            has: () => this.hasStick(arg),
            query: () => this.getStickValue(arg),
            direction: () => this.getStickDirection(arg)
        };
    }
    raw() {
        return [...navigator.getGamepads()][this.index] || null;
    }
    buttonIndex(arg) {
        const raw = this.raw();
        return raw && lookupButtonIndex(raw, arg);
    }
    stickIndex(arg) {
        const raw = this.raw();
        return raw && lookupStickIndex(raw, arg);
    }
}
class AnyGamepad {
    isSomethingConnected() {
        return !!exports.Controllers.getGamepads().find((c) => !!c && c.isSomethingConnected());
    }
    isRecognized() {
        return !!exports.Controllers.getGamepads().find((c) => !!c && c.isRecognized());
    }
    hasButton(arg) {
        return !!exports.Controllers.getGamepads().find((c) => !!c && c.hasButton(arg));
    }
    isButtonPressed(arg) {
        for (const c of exports.Controllers.getGamepads()) {
            if (c && c.isButtonPressed(arg)) {
                return true;
            }
        }
        return false;
    }
    button(arg) {
        return {
            has: () => this.hasButton(arg),
            query: () => this.isButtonPressed(arg)
        };
    }
    hasStick(arg) {
        return !!exports.Controllers.getGamepads().find((c) => !!c && c.hasStick(arg));
    }
    getStickValue(arg) {
        let xSum = 0;
        let ySum = 0;
        for (const c of exports.Controllers.getGamepads()) {
            if (c) {
                const { x, y } = c.getStickValue(arg);
                xSum += x;
                ySum += y;
            }
        }
        return { x: xSum, y: ySum };
    }
    getStickDirection(arg) {
        const gamepad = exports.Controllers.getGamepads().find((c) => !!c && c.getStickDirection(arg) !== null) || null;
        return gamepad && gamepad.getStickDirection(arg);
    }
    stick(arg) {
        return {
            has: () => this.hasStick(arg),
            query: () => this.getStickValue(arg),
            direction: () => this.getStickDirection(arg)
        };
    }
}
exports.Controllers = new (class ControllersSingleton {
    constructor() {
        this.cache = [];
        this.any = null;
    }
    getGamepads() {
        return [...navigator.getGamepads()].map((raw, index) => raw && this.getGamepad(index));
    }
    getGamepad(index) {
        this.cache[index] = this.cache[index] || new ControllerGamepad(index);
        return this.cache[index];
    }
    getAnyGamepad() {
        if (this.any) {
            return this.any;
        }
        const anyGamepad = new AnyGamepad();
        this.any = anyGamepad;
        return anyGamepad;
    }
    key(key, context, modifiers) {
        const mods = new Set(modifiers ? modifiers : []);
        let pressed = false;
        const checkModifiers = (m, evt) => {
            return m.has(KEY_MODIFIER.ALT) === evt.altKey &&
                m.has(KEY_MODIFIER.CTRL) === evt.ctrlKey &&
                m.has(KEY_MODIFIER.META) === evt.metaKey &&
                m.has(KEY_MODIFIER.SHIFT) === evt.shiftKey;
        };
        const checkActiveElement = (root, current) => {
            if (root === current) {
                return true;
            }
            if (current !== null && current.parentElement) {
                return checkActiveElement(root, current.parentElement);
            }
            return false;
        };
        const onKeyDown = (evt) => {
            if (evt.key === key && checkModifiers(mods, evt)) {
                if (!context || checkActiveElement(context, window.document.activeElement)) {
                    pressed = true;
                    evt.preventDefault();
                }
            }
        };
        const onKeyUp = (evt) => {
            if (evt.key === key && checkModifiers(mods, evt)) {
                if (!context || checkActiveElement(context, window.document.activeElement)) {
                    pressed = false;
                    evt.preventDefault();
                }
            }
        };
        window.addEventListener('keydown', onKeyDown);
        window.addEventListener('keyup', onKeyUp);
        return {
            has() { return true; },
            query: () => pressed,
            dispose: () => {
                window.removeEventListener('keydown', onKeyDown);
                window.removeEventListener('keyup', onKeyUp);
            }
        };
    }
    or(arg) {
        return {
            has() { return !!arg.find((b) => b.has()); },
            query() { return !!arg.find((b) => b.query()); }
        };
    }
    asStick(up, down, left, right) {
        function query() {
            const u = up.query();
            const d = down.query();
            const l = left.query();
            const r = right.query();
            const x = 0 + (u ? -1 : 0) + (d ? 1 : 0);
            const y = 0 + (l ? -1 : 0) + (r ? 1 : 0);
            return { x, y };
        }
        return {
            query,
            has() {
                return up.has() && down.has() && left.has() && right.has();
            },
            direction() {
                const { x, y } = query();
                if (x < 0 && y === 0) {
                    return DIRECTION.UP;
                }
                if (x > 0 && y === 0) {
                    return DIRECTION.DOWN;
                }
                if (y < 0 && x === 0) {
                    return DIRECTION.LEFT;
                }
                if (y > 0 && x === 0) {
                    return DIRECTION.RIGHT;
                }
                return null;
            }
        };
    }
    orSticks(arg) {
        function query() {
            let xSum = 0;
            let ySum = 0;
            for (const stick of arg) {
                const { x, y } = stick.query();
                xSum += x;
                ySum += y;
            }
            return { x: xSum, y: ySum };
        }
        return {
            query,
            has() {
                return !!arg.find((c) => c.has());
            },
            direction() {
                const { x, y } = query();
                if (x < 0 && y === 0) {
                    return DIRECTION.UP;
                }
                if (x > 0 && y === 0) {
                    return DIRECTION.DOWN;
                }
                if (y < 0 && x === 0) {
                    return DIRECTION.LEFT;
                }
                if (y > 0 && x === 0) {
                    return DIRECTION.RIGHT;
                }
                return null;
            }
        };
    }
})(); // new Singleton
exports.arrowKeysAsStick = (root) => exports.Controllers.asStick(exports.Controllers.key('ArrowUp', root), exports.Controllers.key('ArrowDown', root), exports.Controllers.key('ArrowLeft', root), exports.Controllers.key('ArrowRight', root));
exports.wsadKeysAsStick = (root) => exports.Controllers.asStick(exports.Controllers.key('w', root), exports.Controllers.key('s', root), exports.Controllers.key('a', root), exports.Controllers.key('d', root));
exports.dpadAsStick = (arg) => exports.Controllers.asStick(arg.button(BUTTON_TYPE.ARROW_UP), arg.button(BUTTON_TYPE.ARROW_DOWN), arg.button(BUTTON_TYPE.ARROW_LEFT), arg.button(BUTTON_TYPE.ARROW_RIGHT));
//# sourceMappingURL=controller.js.map