"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
require("babel-polyfill"); // tslint:disable-line:no-implicit-dependencies
const engine_1 = require("./engine");
const rule_1 = require("./models/rule");
const parser_1 = __importDefault(require("./parser/parser"));
const serializer_1 = __importDefault(require("./parser/serializer"));
const util_1 = require("./util");
let currentEngine = null;
let gameLoop = null;
let awaitingMessage = false;
let receivedMessage = false;
let lastTick = 0;
onmessage = (event) => {
    const msg = event.data;
    switch (msg.type) {
        case util_1.MESSAGE_TYPE.ON_GAME_CHANGE:
            loadGame(msg.code, msg.level, msg.checkpoint);
            break;
        case util_1.MESSAGE_TYPE.PAUSE:
            postMessage({ type: msg.type, payload: pauseGame() });
            break;
        case util_1.MESSAGE_TYPE.RESUME:
            postMessage({ type: msg.type, payload: resumeGame() });
            break;
        case util_1.MESSAGE_TYPE.PRESS:
            postMessage({ type: msg.type, payload: press(msg.button) });
            break;
        case util_1.MESSAGE_TYPE.CLOSE:
            postMessage({ type: msg.type, payload: closeGame() });
            break;
        case util_1.MESSAGE_TYPE.ON_MESSAGE_DONE:
            receivedMessage = true;
            break;
        default:
            throw new Error(`ERROR: Unsupported webworker message type "${JSON.stringify(event.data)}"`);
    }
};
const getEngine = () => {
    if (!currentEngine) {
        throw new Error(`Game has not been loaded yet`);
    }
    return currentEngine;
};
// This uses setTimeout because when a MESSAGE_DONE is received, we need to resume the game
const runPlayLoop = () => __awaiter(this, void 0, void 0, function* () {
    if (gameLoop) {
        clearTimeout(gameLoop);
        gameLoop = null;
    }
    if (awaitingMessage) {
        return;
    }
    if (util_1.shouldTick(getEngine().getGameData().metadata, lastTick)) {
        lastTick = Date.now();
        yield tick();
    }
    gameLoop = setTimeout(runPlayLoop, 20);
});
let previousMessage = ''; // a dev-invariant checker that ensures we do not show the same message twice
class Handler {
    onGameChange(gameData) {
        postMessage({ type: util_1.MESSAGE_TYPE.ON_GAME_CHANGE, payload: (new serializer_1.default(gameData)).toJson() });
    }
    onPress(dir) {
        if (!dir) {
            throw new Error(`BUG: No direction provided to onPress`);
        }
        postMessage({ type: util_1.MESSAGE_TYPE.ON_PRESS, direction: dir });
    }
    onMessage(msg) {
        return __awaiter(this, void 0, void 0, function* () {
            if (previousMessage === msg) {
                throw new Error(`BUG: Should not show the same message twice. "${msg}"`);
            }
            previousMessage = msg;
            // Wait until the user dismissed the message
            if (awaitingMessage) {
                throw new Error(`BUG: should not already be awaiting a message`);
            }
            awaitingMessage = true;
            receivedMessage = false;
            pauseGame();
            postMessage({ type: util_1.MESSAGE_TYPE.ON_MESSAGE, message: msg });
            yield util_1.pollingPromise(50, () => receivedMessage);
            awaitingMessage = false;
            // resumeGame() No need to resume since we are inside a `await tick()` and at the end of it it will start back up (via a call to setTimeout)
        });
    }
    onLevelChange(level, cells, message) {
        let newCells = null;
        if (cells) {
            newCells = cells.map((row) => toCellsJson(row));
        }
        postMessage({ type: util_1.MESSAGE_TYPE.ON_LEVEL_CHANGE, level, cells: newCells, message });
    }
    onWin() {
        postMessage({ type: util_1.MESSAGE_TYPE.ON_WIN });
    }
    onSound(sound) {
        return __awaiter(this, void 0, void 0, function* () {
            postMessage({ type: util_1.MESSAGE_TYPE.ON_SOUND, soundCode: sound.soundCode });
        });
    }
    onTick(changedCells, checkpoint, hasAgain, a11yMessages) {
        postMessage({ type: util_1.MESSAGE_TYPE.ON_TICK, changedCells: toCellsJson(changedCells), checkpoint, hasAgain, a11yMessages: a11yMessages.map(toA11yMessageJson) });
    }
    onPause() {
        postMessage({ type: util_1.MESSAGE_TYPE.ON_PAUSE });
    }
    onResume() {
        postMessage({ type: util_1.MESSAGE_TYPE.ON_RESUME });
    }
}
const loadGame = (code, level, checkpoint) => {
    pauseGame();
    previousMessage = ''; // clear this dev-invariant-tester field since it is a new game
    const { data } = parser_1.default.parse(code);
    postMessage({ type: util_1.MESSAGE_TYPE.ON_GAME_CHANGE, payload: (new serializer_1.default(data)).toJson() });
    currentEngine = new engine_1.GameEngine(data, new Handler());
    currentEngine.setLevel(level, checkpoint);
    runPlayLoop(); // tslint:disable-line:no-floating-promises
};
const pauseGame = () => {
    if (gameLoop !== null) {
        clearTimeout(gameLoop);
        gameLoop = null;
        postMessage({ type: util_1.MESSAGE_TYPE.ON_PAUSE });
    }
};
const resumeGame = () => {
    postMessage({ type: util_1.MESSAGE_TYPE.ON_RESUME });
    runPlayLoop(); // tslint:disable-line:no-floating-promises
};
const tick = () => __awaiter(this, void 0, void 0, function* () {
    const engine = getEngine();
    const { changedCells, didWinGame, didLevelChange, wasAgainTick } = yield engine.tick();
    // Response needs to be serializable
    return {
        changedCells: toCellsJson(changedCells),
        didWinGame,
        didLevelChange,
        wasAgainTick
    };
});
const toCellsJson = (cells) => {
    return [...cells].map(toCellJson);
};
const toCellJson = (cell) => {
    const { colIndex, rowIndex } = cell;
    return {
        colIndex,
        rowIndex,
        spriteNames: cell.getSprites().map((sprite) => sprite.getName())
    };
};
const toA11yMessageJson = (message) => {
    switch (message.type) {
        case rule_1.A11Y_MESSAGE_TYPE.ADD:
            return Object.assign({}, message, { cell: toCellJson(message.cell), sprites: [...message.sprites].map(toSpriteName) });
        case rule_1.A11Y_MESSAGE_TYPE.MOVE:
            return Object.assign({}, message, { oldCell: toCellJson(message.oldCell), newCell: toCellJson(message.newCell), sprite: toSpriteName(message.sprite) });
        case rule_1.A11Y_MESSAGE_TYPE.REMOVE:
            return Object.assign({}, message, { cell: toCellJson(message.cell), sprites: [...message.sprites].map(toSpriteName) });
        case rule_1.A11Y_MESSAGE_TYPE.REPLACE:
            return Object.assign({}, message, { cell: toCellJson(message.cell), replacements: [...message.replacements].map(({ oldSprite, newSprite }) => ({ oldSprite: toSpriteName(oldSprite), newSprite: toSpriteName(newSprite) })) });
    }
};
const toSpriteName = (sprite) => {
    return sprite.getName();
};
const press = (button) => {
    return getEngine().press(button);
};
const closeGame = () => {
    pauseGame();
    currentEngine = null;
};
//# sourceMappingURL=index-webworker.js.map